/*
 * Copyright  2012 Ilkka Lehtoranta
 *
 * $Id: methodstack.c,v 1.1 2012/12/21 23:30:15 itix Exp $
 */

#include <dos/dosextens.h>
#include <exec/execbase.h>
#include <proto/alib.h>
#include <proto/exec.h>
#include <proto/intuition.h>

#include "../../development/os-include/private/vapor/vapor.h"

STATIC struct SignalSemaphore	semaphore;
STATIC struct MinList list = { (APTR)&list.mlh_Tail, NULL, (APTR)&list };
STATIC struct Task *GUIThread;

enum
{
	ct_method,
	ct_methodsync,
	ct_set,
};

typedef struct
{
	struct Message Message;
	ULONG calltype;
	ULONG result;
	APTR  obj;
	ULONG m[0];
} method;

VOID methodstack_init()
{
	GUIThread = SysBase->ThisTask;
	InitSemaphore(&semaphore);
}

VOID methodstack_cleanup()
{
	APTR node, next;

	ITERATELISTSAFE(node, next, &list)
	{
		FreeVec(node);
	}
}

STATIC VOID methodstack_add(method *m)
{
	ObtainSemaphore(&semaphore);
	ADDTAIL(&list, m);
	ReleaseSemaphore(&semaphore);
	Signal((struct Task *)GUIThread, SIGBREAKF_CTRL_E);
}

VOID methodstack_push(APTR obj, ULONG cnt, ...)
{
	method *m;
	va_list va;
	va_start(va, cnt);

	if ((m = AllocVec(sizeof(*m) + cnt * sizeof(ULONG), MEMF_ANY)))
	{
		ULONG i = 0;

		while (cnt--)
		{
			m->m[i] = va_arg(va, ULONG);
			i++;
		}

		m->calltype = ct_method;
		m->obj = obj;

		methodstack_add(m);
	}

	va_end(va);
}

VOID methodstack_set(APTR obj, ULONG cnt, ...)
{
	method *m;
	va_list va;
	va_start(va, cnt);

	if ((m = AllocVec(sizeof(*m) + (cnt + 1) * sizeof(ULONG), MEMF_ANY)))
	{
		ULONG i = 0;

		while (cnt--)
		{
			m->m[i] = va_arg(va, ULONG);
			i++;
		}

		m->m[i] = TAG_DONE;

		if (SysBase->ThisTask == GUIThread)
		{
			SetAttrsA(obj, (struct TagItem *)&m->m[0]);
		}
		else
		{
			m->calltype = ct_set;
			m->obj = obj;

			methodstack_add(m);
		}
	}

	va_end(va);
}

ULONG methodstack_pushsync(APTR obj, ULONG cnt, ...)
{
	method *m;
	ULONG res;
	va_list va;
	va_start(va, cnt);

	m = AllocVec(sizeof(*m) + cnt * sizeof(ULONG), MEMF_ANY);
	res = 0;

	if (m)
	{
		struct Process *thisproc = (APTR)SysBase->ThisTask;
		ULONG i = 0;

		while (cnt--)
		{
			m->m[i] = va_arg(va, ULONG);
			i++;
		}

		if (thisproc == (APTR)GUIThread)
		{
			res = DoMethodA(obj, (Msg)&m->m[0]);
		}
		else
		{
			struct MsgPort *replyport = &thisproc->pr_MsgPort;

			m->Message.mn_ReplyPort = replyport;
			m->calltype = ct_methodsync;
			m->obj = obj;

			methodstack_add(m);

			WaitPort(replyport);
			GetMsg(replyport);
			res = m->result;
		}

		FreeVec(m);
	}

	va_end(va);

	return (res);
}

VOID methodstack_check()
{
	ULONG	empty;

	do
	{
		method *m;

		ObtainSemaphore(&semaphore);
		m = REMHEAD(&list);
		empty = IsListEmpty((struct List *)&list);
		ReleaseSemaphore(&semaphore);

		if (!m)
			break;

		if (m->calltype == ct_set)
			SetAttrsA(m->obj, (struct TagItem *)&m->m[0]);
		else
			m->result = DoMethodA(m->obj, (Msg)&m->m[0]);

		if (m->calltype == ct_methodsync)
			ReplyMsg(&m->Message);
		else
			FreeVec(m);
	}
	while (!empty);
}
